@@ -37,7 +37,7 @@
 #include "getdef.h"
 #include "prototypes.h"
 
-#ident "$Id: console.c 2130 2008-06-13 18:11:09Z nekral-guest $"
+#ident "$Id: console.c 2790 2009-04-24 23:03:14Z nekral-guest $"
 
 /* local function prototypes */
 static bool is_listed (const char *cfgin, const char *tty, bool def);
@@ -68,8 +68,8 @@ static bool is_listed (const char *cfgin, const char *tty, bool def)
 	 */
 
 	if (*cons != '/') {
-		cons = strcpy (buf, cons);
-		while ((s = strtok (cons, ":")) != NULL) {
+		strcpy (buf, cons);
+		while ((s = strtok (buf, ":")) != NULL) {
 			if (strcmp (s, tty) == 0) {
 				return true;
 			}
