@@ -32,8 +32,9 @@
 
 #include <config.h>
 
-#ident "$Id: copydir.c 2643 2009-04-11 15:34:10Z nekral-guest $"
+#ident "$Id: copydir.c 2900 2009-05-10 18:21:58Z nekral-guest $"
 
+#include <assert.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/time.h>
@@ -44,17 +45,17 @@
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
 #endif
-static const char *src_orig;
-static const char *dst_orig;
+static /*@null@*/const char *src_orig;
+static /*@null@*/const char *dst_orig;
 
 struct link_name {
 	dev_t ln_dev;
 	ino_t ln_ino;
-	int ln_count;
+	nlink_t ln_count;
 	char *ln_name;
-	struct link_name *ln_next;
+	/*@dependent@*/struct link_name *ln_next;
 };
-static struct link_name *links;
+static /*@exposed@*/struct link_name *links;
 
 static int copy_entry (const char *src, const char *dst,
                        long int uid, long int gid);
@@ -62,6 +63,7 @@ static int copy_dir (const char *src, const char *dst,
                      const struct stat *statp, const struct timeval mt[],
                      long int uid, long int gid);
 #ifdef	S_IFLNK
+static char *readlink_malloc (const char *filename);
 static int copy_symlink (const char *src, const char *dst,
                          const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid);
@@ -120,7 +122,7 @@ int selinux_file_context (const char *dst_name)
 /*
  * remove_link - delete a link from the linked list
  */
-static void remove_link (struct link_name *ln)
+static void remove_link (/*@only@*/struct link_name *ln)
 {
 	struct link_name *lp;
 
@@ -137,6 +139,8 @@ static void remove_link (struct link_name *ln)
 	}
 
 	if (NULL == lp) {
+		free (ln->ln_name);
+		free (ln);
 		return;
 	}
 
@@ -149,7 +153,7 @@ static void remove_link (struct link_name *ln)
  * check_link - see if a file is really a link
  */
 
-static struct link_name *check_link (const char *name, const struct stat *sb)
+static /*@exposed@*/ /*@null@*/struct link_name *check_link (const char *name, const struct stat *sb)
 {
 	struct link_name *lp;
 	size_t src_len;
@@ -157,7 +161,11 @@ static struct link_name *check_link (const char *name, const struct stat *sb)
 	size_t name_len;
 	size_t len;
 
-	for (lp = links; lp; lp = lp->ln_next) {
+	/* copy_tree () must be the entry point */
+	assert (NULL != src_orig);
+	assert (NULL != dst_orig);
+
+	for (lp = links; NULL != lp; lp = lp->ln_next) {
 		if ((lp->ln_dev == sb->st_dev) && (lp->ln_ino == sb->st_ino)) {
 			return lp;
 		}
@@ -192,8 +200,6 @@ static struct link_name *check_link (const char *name, const struct stat *sb)
 int copy_tree (const char *src_root, const char *dst_root,
                long int uid, long int gid)
 {
-	char src_name[1024];
-	char dst_name[1024];
 	int err = 0;
 	bool set_orig = false;
 	struct DIRECT *ent;
@@ -233,27 +239,36 @@ int copy_tree (const char *src_root, const char *dst_root,
 		 */
 		if ((strcmp (ent->d_name, ".") != 0) &&
 		    (strcmp (ent->d_name, "..") != 0)) {
-			/*
-			 * Make sure the resulting source and destination
-			 * filenames will fit in their buffers.
-			 */
-			if (   (strlen (src_root) + strlen (ent->d_name) + 2 >
-			        sizeof src_name)
-			    || (strlen (dst_root) + strlen (ent->d_name) + 2 >
-			        sizeof dst_name)) {
+			char *src_name;
+			char *dst_name;
+			size_t src_len = strlen (ent->d_name) + 2;
+			size_t dst_len = strlen (ent->d_name) + 2;
+			src_len += strlen (src_root);
+			dst_len += strlen (dst_root);
+
+			src_name = (char *) malloc (src_len);
+			dst_name = (char *) malloc (dst_len);
+
+			if ((NULL == src_name) || (NULL == dst_name)) {
 				err = -1;
 			} else {
 				/*
 				 * Build the filename for both the source and
 				 * the destination files.
 				 */
-				snprintf (src_name, sizeof src_name, "%s/%s",
+				snprintf (src_name, src_len, "%s/%s",
 				          src_root, ent->d_name);
-				snprintf (dst_name, sizeof dst_name, "%s/%s",
+				snprintf (dst_name, dst_len, "%s/%s",
 				          dst_root, ent->d_name);
 
 				err = copy_entry (src_name, dst_name, uid, gid);
 			}
+			if (NULL != src_name) {
+				free (src_name);
+			}
+			if (NULL != dst_name) {
+				free (dst_name);
+			}
 		}
 	}
 	(void) closedir (dir);
@@ -268,6 +283,10 @@ int copy_tree (const char *src_root, const char *dst_root,
 	setfscreatecon (NULL);
 #endif
 
+	/* FIXME: with the call to remove_link, we could also check that
+	 *        no links remain in links.
+	 * assert (NULL == links); */
+
 	return err;
 }
 
@@ -404,6 +423,41 @@ static int copy_dir (const char *src, const char *dst,
 }
 
 #ifdef	S_IFLNK
+/*
+ * readlink_malloc - wrapper for readlink
+ *
+ * return NULL on error.
+ * The return string shall be freed by the caller.
+ */
+static char *readlink_malloc (const char *filename)
+{
+	size_t size = 1024;
+
+	while (1) {
+		ssize_t nchars;
+		char *buffer = (char *) malloc (size);
+		if (NULL == buffer) {
+			return NULL;
+		}
+
+		nchars = readlink (filename, buffer, size);
+
+		if (nchars < 0) {
+			return NULL;
+		}
+
+		if ( (size_t) nchars < size) { /* The buffer was large enough */
+			/* readlink does not nul-terminate */
+			buffer[nchars] = '\0';
+			return buffer;
+		}
+
+		/* Try again with a bigger buffer */
+		free (buffer);
+		size *= 2;
+	}
+}
+
 /*
  * copy_symlink - copy a symlink
  *
@@ -418,10 +472,11 @@ static int copy_symlink (const char *src, const char *dst,
                          const struct stat *statp, const struct timeval mt[],
                          long int uid, long int gid)
 {
-	char oldlink[1024];
-	char dummy[1024];
-	int len;
-	int err = 0;
+	char *oldlink;
+
+	/* copy_tree () must be the entry point */
+	assert (NULL != src_orig);
+	assert (NULL != dst_orig);
 
 	/*
 	 * Get the name of the file which the link points
@@ -431,17 +486,25 @@ static int copy_symlink (const char *src, const char *dst,
 	 * destination directory name.
 	 */
 
-	len = readlink (src, oldlink, sizeof (oldlink) - 1);
-	if (len < 0) {
+	oldlink = readlink_malloc (src);
+	if (NULL == oldlink) {
 		return -1;
 	}
-	oldlink[len] = '\0';	/* readlink() does not NUL-terminate */
+
+	/* If src was a link to an entry of the src_orig directory itself,
+	 * create a link to the corresponding entry in the dst_orig
+	 * directory.
+	 */
 	if (strncmp (oldlink, src_orig, strlen (src_orig)) == 0) {
-		snprintf (dummy, sizeof dummy, "%s%s",
+		size_t len = strlen (dst_orig) + strlen (oldlink) - strlen (src_orig) + 1;
+		char *dummy = (char *) malloc (len);
+		snprintf (dummy, len, "%s%s",
 		          dst_orig,
 		          oldlink + strlen (src_orig));
-		strcpy (oldlink, dummy);
+		free (oldlink);
+		oldlink = dummy;
 	}
+
 #ifdef WITH_SELINUX
 	selinux_file_context (dst);
 #endif
@@ -449,8 +512,10 @@ static int copy_symlink (const char *src, const char *dst,
 	    || (lchown (dst,
 	                (uid == -1) ? statp->st_uid : (uid_t) uid,
 	                (gid == -1) ? statp->st_gid : (gid_t) gid) != 0)) {
+		free (oldlink);
 		return -1;
 	}
+	free (oldlink);
 
 #ifdef HAVE_LUTIMES
 	/* 2007-10-18: We don't care about
@@ -461,7 +526,7 @@ static int copy_symlink (const char *src, const char *dst,
 	lutimes (dst, mt);
 #endif
 
-	return err;
+	return 0;
 }
 #endif
 
@@ -480,10 +545,13 @@ static int copy_hardlink (const char *src, const char *dst,
 	if (link (lp->ln_name, dst) != 0) {
 		return -1;
 	}
+
+	/* FIXME: why is it unlinked? This is a copy, not a move*/
 	if (unlink (src) != 0) {
 		return -1;
 	}
 
+	/* FIXME: idem, although it may never be used again */
 	/* If the file could be unlinked, decrement the links counter,
 	 * and delete the file if it was the last reference */
 	lp->ln_count--;
@@ -553,12 +621,12 @@ static int copy_file (const char *src, const char *dst,
 #ifdef WITH_SELINUX
 	selinux_file_context (dst);
 #endif
-	ofd = open (dst, O_WRONLY | O_CREAT | O_TRUNC, 0);
+	ofd = open (dst, O_WRONLY | O_CREAT | O_TRUNC, statp->st_mode & 07777);
 	if (   (ofd < 0)
-	    || (chown (dst,
-	               (uid == -1) ? statp->st_uid : (uid_t) uid,
-	               (gid == -1) ? statp->st_gid : (gid_t) gid) != 0)
-	    || (chmod (dst, statp->st_mode & 07777) != 0)) {
+	    || (fchown (ofd,
+	                (uid == -1) ? statp->st_uid : (uid_t) uid,
+	                (gid == -1) ? statp->st_gid : (gid_t) gid) != 0)
+	    || (fchmod (ofd, statp->st_mode & 07777) != 0)) {
 		(void) close (ifd);
 		return -1;
 	}
@@ -600,7 +668,7 @@ static int copy_file (const char *src, const char *dst,
 
 int remove_tree (const char *root)
 {
-	char new_name[1024];
+	char *new_name = NULL;
 	int err = 0;
 	struct DIRECT *ent;
 	struct stat sb;
@@ -627,6 +695,7 @@ int remove_tree (const char *root)
 	}
 
 	while ((ent = readdir (dir))) {
+		size_t new_len = strlen (root) + strlen (ent->d_name) + 2;
 
 		/*
 		 * Skip the "." and ".." entries
@@ -641,12 +710,15 @@ int remove_tree (const char *root)
 		 * Make the filename for the current entry.
 		 */
 
-		if (strlen (root) + strlen (ent->d_name) + 2 > sizeof new_name) {
+		if (NULL != new_name) {
+			free (new_name);
+		}
+		new_name = (char *) malloc (new_len);
+		if (NULL == new_name) {
 			err = -1;
 			break;
 		}
-		snprintf (new_name, sizeof new_name, "%s/%s", root,
-			  ent->d_name);
+		snprintf (new_name, new_len, "%s/%s", root, ent->d_name);
 		if (LSTAT (new_name, &sb) == -1) {
 			continue;
 		}
@@ -669,6 +741,9 @@ int remove_tree (const char *root)
 			}
 		}
 	}
+	if (NULL != new_name) {
+		free (new_name);
+	}
 	(void) closedir (dir);
 
 	if (0 == err) {
