@@ -1,5 +1,7 @@
 /*
- * Copyright 1989 - 1991, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1991, Julianne Frances Haugh
+ * Copyright (c) 1996 - 1998, Marek Michałkiewicz
+ * Copyright (c) 2003 - 2006, Tomasz Kłoczko
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,26 +12,26 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: shell.c,v 1.11 2005/08/31 17:24:58 kloczek Exp $"
+#ident "$Id: shell.c 2760 2009-04-22 21:21:14Z nekral-guest $"
 
 #include <stdio.h>
 #include <errno.h>
@@ -51,13 +53,15 @@ extern size_t newenvc;
  *	the file.  If all that fails, give up in disgust ...
  */
 
-void shell (const char *file, const char *arg)
+int shell (const char *file, /*@null@*/const char *arg, char *const envp[])
 {
 	char arg0[1024];
 	int err;
 
-	if (file == (char *) 0)
-		exit (1);
+	if (file == (char *) 0) {
+		errno = EINVAL;
+		return errno;
+	}
 
 	/*
 	 * The argv[0]'th entry is usually the path name, but
@@ -65,22 +69,17 @@ void shell (const char *file, const char *arg)
 	 * that.  So, we determine the 0'th entry only if they
 	 * don't want to tell us what it is themselves.
 	 */
-
 	if (arg == (char *) 0) {
 		snprintf (arg0, sizeof arg0, "-%s", Basename ((char *) file));
 		arg = arg0;
 	}
-#ifdef DEBUG
-	printf ("Executing shell %s\n", file);
-#endif
 
 	/*
 	 * First we try the direct approach.  The system should be
 	 * able to figure out what we are up to without too much
 	 * grief.
 	 */
-
-	execle (file, arg, (char *) 0, newenvp);
+	execle (file, arg, (char *) 0, envp);
 	err = errno;
 
 	/* Linux handles #! in the kernel, and bash doesn't make
@@ -92,18 +91,18 @@ void shell (const char *file, const char *arg)
 	 * relies on the standard shell being able to make sense
 	 * of the "#!" magic number.
 	 */
-
 	if (err == ENOEXEC) {
 		FILE *fp;
 
-		if ((fp = fopen (file, "r"))) {
+		fp = fopen (file, "r");
+		if (NULL != fp) {
 			if (getc (fp) == '#' && getc (fp) == '!') {
-				fclose (fp);
+				(void) fclose (fp);
 				execle ("/bin/sh", "sh",
-					file, (char *) 0, newenvp);
+				        file, (char *) 0, envp);
 				err = errno;
 			} else {
-				fclose (fp);
+				(void) fclose (fp);
 			}
 		}
 	}
@@ -114,9 +113,8 @@ void shell (const char *file, const char *arg)
 	 * how to execute this stupid shell, so I might as well give
 	 * up in disgust ...
 	 */
-
 	snprintf (arg0, sizeof arg0, _("Cannot execute %s"), file);
 	errno = err;
 	perror (arg0);
-	exit (1);
+	return err;
 }
