@@ -32,16 +32,19 @@
 
 #include <config.h>
 
-#ident "$Id: newgrp.c 2348 2008-09-06 12:51:53Z nekral-guest $"
+#ident "$Id: newgrp.c 2851 2009-04-30 21:39:38Z nekral-guest $"
 
 #include <errno.h>
 #include <grp.h>
 #include <pwd.h>
 #include <stdio.h>
+#include <assert.h>
 #include "defines.h"
 #include "getdef.h"
 #include "prototypes.h"
+/*@-exitarg@*/
 #include "exitcodes.h"
+
 /*
  * Global variables
  */
@@ -52,13 +55,13 @@ extern char **environ;
 
 #ifdef HAVE_SETGROUPS
 static int ngroups;
-static GETGROUPS_T *grouplist;
+static /*@null@*/ /*@only@*/GETGROUPS_T *grouplist;
 #endif
 
 static bool is_newgrp;
 
 #ifdef WITH_AUDIT
-char audit_buf[80];
+static char audit_buf[80];
 #endif
 
 /* local function prototypes */
@@ -74,9 +77,9 @@ static void syslog_sg (const char *name, const char *group);
 static void usage (void)
 {
 	if (is_newgrp) {
-		fputs (_("Usage: newgrp [-] [group]\n"), stderr);
+		(void) fputs (_("Usage: newgrp [-] [group]\n"), stderr);
 	} else {
-		fputs (_("Usage: sg group [[-c] command]\n"), stderr);
+		(void) fputs (_("Usage: sg group [[-c] command]\n"), stderr);
 	}
 }
 
@@ -84,7 +87,7 @@ static void usage (void)
  * find_matching_group - search all groups of a given group id for
  *                       membership of a given username
  */
-static struct group *find_matching_group (const char *name, gid_t gid)
+static /*@null@*/struct group *find_matching_group (const char *name, gid_t gid)
 {
 	struct group *gr;
 	char **look;
@@ -194,8 +197,8 @@ static void check_perms (const struct group *grp,
 			SYSLOG ((LOG_INFO,
 				 "Invalid password for group '%s' from '%s'",
 				 groupname, pwd->pw_name));
-			sleep (1);
-			fputs (_("Invalid password.\n"), stderr);
+			(void) sleep (1);
+			(void) fputs (_("Invalid password.\n"), stderr);
 			goto failure;
 		}
 #ifdef WITH_AUDIT
@@ -228,7 +231,7 @@ failure:
 		              (unsigned int) getuid (), 0);
 	}
 #endif
-	exit (1);
+	exit (EXIT_FAILURE);
 }
 
 #ifdef USE_SYSLOG
@@ -307,7 +310,7 @@ static void syslog_sg (const char *name, const char *group)
 				              (unsigned int) getuid (), 0);
 			}
 #endif
-			exit (1);
+			exit (EXIT_FAILURE);
 		} else if (child != 0) {
 			/* parent - wait for child to finish, then log session close */
 			int cst = 0;
@@ -344,7 +347,7 @@ static void syslog_sg (const char *name, const char *group)
 				         (unsigned long) gid, name));
 			}
 			closelog ();
-			exit (0);
+			exit (E_SUCCESS);
 		}
 
 		/* child - restore signals to their default state */
@@ -375,7 +378,7 @@ int main (int argc, char **argv)
 	char *command = NULL;
 	char **envp = environ;
 	struct passwd *pwd;
-	struct group *grp;
+	/*@null@*/struct group *grp;
 
 #ifdef SHADOWGRP
 	struct sgrp *sgrp;
@@ -432,7 +435,7 @@ int main (int argc, char **argv)
 		SYSLOG ((LOG_WARN, "Cannot determine the user name of the caller (UID %lu)",
 		         (unsigned long) getuid ()));
 		closelog ();
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 	name = pwd->pw_name;
 
@@ -472,7 +475,7 @@ int main (int argc, char **argv)
 		} else {
 			usage ();
 			closelog ();
-			exit (1);
+			exit (EXIT_FAILURE);
 		}
 		if (argc > 0) {
 
@@ -554,7 +557,7 @@ int main (int argc, char **argv)
 			              (unsigned int) getuid (), 0);
 		}
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 #endif				/* HAVE_SETGROUPS */
 
@@ -619,6 +622,7 @@ int main (int argc, char **argv)
 		 * Re-read the group entry for further processing.
 		 */
 		grp = xgetgrnam (group);
+		assert (NULL != grp);
 	}
 #ifdef SHADOWGRP
 	sgrp = getsgnam (group);
@@ -659,7 +663,7 @@ int main (int argc, char **argv)
 	}
 	if (i == ngroups) {
 		if (ngroups >= sysconf (_SC_NGROUPS_MAX)) {
-			fputs (_("too many groups\n"), stderr);
+			(void) fputs (_("too many groups\n"), stderr);
 		} else {
 			grouplist[ngroups++] = gid;
 			if (setgroups (ngroups, grouplist) != 0) {
@@ -669,6 +673,19 @@ int main (int argc, char **argv)
 	}
 #endif
 
+	/*
+	 * Close all files before changing the user/group IDs.
+	 *
+	 * The needed structure should have been copied before, or
+	 * permission to read the database will be required.
+	 */
+	endspent ();
+#ifdef	SHADOWGRP
+	endsgent ();
+#endif
+	endpwent ();
+	endgrent ();
+
 	/*
 	 * Set the effective GID to the new group id and the effective UID
 	 * to the real UID. For root, this also sets the real GID to the
@@ -683,7 +700,7 @@ int main (int argc, char **argv)
 		              audit_buf, NULL,
 		              (unsigned int) getuid (), 0);
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	if (setuid (getuid ()) != 0) {
@@ -695,7 +712,7 @@ int main (int argc, char **argv)
 		              audit_buf, NULL,
 		              (unsigned int) getuid (), 0);
 #endif
-		exit (1);
+		exit (EXIT_FAILURE);
 	}
 
 	/*
@@ -713,7 +730,7 @@ int main (int argc, char **argv)
 		              (unsigned int) getuid (), 0);
 #endif
 		perror ("/bin/sh");
-		exit (errno == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
+		exit ((errno == ENOENT) ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
 	}
 
 	/*
@@ -748,13 +765,6 @@ int main (int argc, char **argv)
 	 */
 	cp = Basename ((char *) prog);
 
-	endspent ();
-#ifdef	SHADOWGRP
-	endsgent ();
-#endif
-	endpwent ();
-	endgrent ();
-
 	/*
 	 * Switch back to her home directory if i am doing login
 	 * initialization.
@@ -792,8 +802,8 @@ int main (int argc, char **argv)
 	 * the previous environment which should be the user's login shell.
 	 */
 	err = shell (prog, initflag ? (char *) 0 : cp, newenvp);
-	exit (err == ENOENT ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
-	/* NOTREACHED */
+	exit ((err == ENOENT) ? E_CMD_NOTFOUND : E_CMD_NOEXEC);
+	/* @notreached@ */
       failure:
 
 	/*
@@ -820,6 +830,6 @@ int main (int argc, char **argv)
 		              (unsigned int) getuid (), 0);
 	}
 #endif
-	exit (1);
+	exit (EXIT_FAILURE);
 }
 
