@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: passwd.c 2612 2009-04-10 22:34:23Z nekral-guest $"
+#ident "$Id: passwd.c 2863 2009-05-09 13:14:31Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -46,7 +46,7 @@
 #include <selinux/flask.h>
 #include <selinux/av_permissions.h>
 #include <selinux/context.h>
-#endif
+#endif				/* WITH_SELINUX */
 #include <time.h>
 #include "defines.h"
 #include "getdef.h"
@@ -55,9 +55,11 @@
 #include "pwauth.h"
 #include "pwio.h"
 #include "shadowio.h"
+
 /*
  * exit status values
  */
+/*@-exitarg@*/
 #define E_SUCCESS	0	/* success */
 #define E_NOPERM	1	/* permission denied */
 #define E_USAGE		2	/* invalid combination of options */
@@ -101,7 +103,7 @@ static long inact = 0;		/* Days without change before locked */
 
 #ifndef USE_PAM
 static bool do_update_age = false;
-#endif
+#endif				/* ! USE_PAM */
 
 static bool pw_locked = false;
 static bool spw_locked = false;
@@ -122,7 +124,7 @@ static bool spw_locked = false;
  */
 static char crypt_passwd[256];
 static bool do_update_pwd = false;
-#endif
+#endif				/* !USE_PAM */
 
 /*
  * External identifiers
@@ -151,7 +153,7 @@ static void update_shadow (void);
 static int check_selinux_access (const char *changed_user,
                                  uid_t changed_uid,
                                  access_vector_t requested_access);
-#endif
+#endif				/* WITH_SELINUX */
 
 /*
  * usage - print command usage and exit
@@ -192,16 +194,16 @@ static int reuse (const char *pass, const struct passwd *pw)
 	const char *FascistHistoryPw (const char *, const struct passwd *);
 
 	reason = FascistHistory (pass, pw);
-#else
+#else				/* !HAVE_LIBCRACK_PW */
 	const char *FascistHistory (const char *, int);
 
 	reason = FascistHistory (pass, pw->pw_uid);
-#endif
+#endif				/* !HAVE_LIBCRACK_PW */
 	if (NULL != reason) {
 		printf (_("Bad password: %s.  "), reason);
 		return 1;
 	}
-#endif
+#endif				/* HAVE_LIBCRACK_HIST */
 	return 0;
 }
 
@@ -223,7 +225,7 @@ static int new_password (const struct passwd *pw)
 
 #ifdef HAVE_LIBCRACK_HIST
 	int HistUpdate (const char *, const char *);
-#endif
+#endif				/* HAVE_LIBCRACK_HIST */
 
 	/*
 	 * Authenticate the user. The user will be prompted for their own
@@ -268,7 +270,7 @@ static int new_password (const struct passwd *pw)
 #ifdef USE_SHA_CRYPT
 		    || (strcmp (method, "SHA256") == 0)
 		    || (strcmp (method, "SHA512") == 0)
-#endif
+#endif				/* USE_SHA_CRYPT */
 		    ) {
 			pass_max_len = -1;
 		} else {
@@ -345,7 +347,7 @@ static int new_password (const struct passwd *pw)
 
 #ifdef HAVE_LIBCRACK_HIST
 	HistUpdate (pw->pw_name, crypt_passwd);
-#endif
+#endif				/* HAVE_LIBCRACK_HIST */
 	STRFCPY (crypt_passwd, cp);
 	return 0;
 }
@@ -434,10 +436,10 @@ static char *date_to_str (time_t t)
 	tm = gmtime (&t);
 #ifdef HAVE_STRFTIME
 	strftime (buf, sizeof buf, "%m/%d/%Y", tm);
-#else
+#else				/* !HAVE_STRFTIME */
 	snprintf (buf, sizeof buf, "%02d/%02d/%04d",
 		  tm->tm_mon + 1, tm->tm_mday, tm->tm_year + 1900);
-#endif
+#endif				/* !HAVE_STRFTIME */
 	return buf;
 }
 
@@ -508,7 +510,7 @@ static char *update_crypt_pw (char *cp)
 	if (do_update_pwd) {
 		cp = insert_crypt_passwd (cp, crypt_passwd);
 	}
-#endif
+#endif				/* !USE_PAM */
 
 	if (dflg) {
 		*cp = '\0';
@@ -517,9 +519,10 @@ static char *update_crypt_pw (char *cp)
 	if (uflg && *cp == '!') {
 		if (cp[1] == '\0') {
 			fprintf (stderr,
-				 _("%s: unlocking the password would result in a passwordless account.\n"
-				   "You should set a password with usermod -p to unlock the password of this account.\n"),
-				 Prog);
+			         _("%s: unlocking the password would result in a passwordless account.\n"
+			           "You should set a password with usermod -p to unlock the password of this account.\n"),
+			         Prog);
+			fail_exit (E_FAILURE);
 		} else {
 			cp++;
 		}
@@ -644,7 +647,7 @@ static void update_shadow (void)
 			nsp->sp_lstchg = -1;
 		}
 	}
-#endif
+#endif				/* !USE_PAM */
 
 	/*
 	 * Force change on next login, like SunOS 4.x passwd -e or Solaris
@@ -723,7 +726,7 @@ static int check_selinux_access (const char *changed_user,
 	return status;
 }
 
-#endif
+#endif				/* WITH_SELINUX */
 
 /*
  * passwd - change a user's password file information
@@ -762,7 +765,7 @@ int main (int argc, char **argv)
 	char *cp;		/* Miscellaneous character pointing  */
 
 	const struct spwd *sp;	/* Shadow file entry for user   */
-#endif
+#endif				/* !USE_PAM */
 
 	(void) setlocale (LC_ALL, "");
 	(void) bindtextdomain (PACKAGE, LOCALEDIR);
@@ -808,9 +811,8 @@ int main (int argc, char **argv)
 			{NULL, 0, NULL, '\0'}
 		};
 
-		while ((c =
-			getopt_long (argc, argv, "adei:kln:qr:Suw:x:",
-				     long_options, &option_index)) != -1) {
+		while ((c = getopt_long (argc, argv, "adei:kln:qr:Suw:x:",
+		                         long_options, &option_index)) != -1) {
 			switch (c) {
 			case 'a':
 				aflg = true;
@@ -991,20 +993,23 @@ int main (int argc, char **argv)
 	   changing a password without entering the old one */
 	if ((is_selinux_enabled() > 0) && (getuid() == 0) &&
 	    (check_selinux_access (name, pw->pw_uid, PASSWD__PASSWD) != 0)) {
-		security_context_t user_context;
-		if (getprevcon(&user_context) < 0) {
-			user_context = strdup("Unknown user context");
+		security_context_t user_context = NULL;
+		const char *user = "Unknown user context";
+		if (getprevcon (&user_context) == 0) {
+			user = user_context;
 		}
-		syslog(LOG_ALERT,
-		       "%s is not authorized to change the password of %s",
-		       user_context, name);
+		SYSLOG ((LOG_ALERT,
+		         "%s is not authorized to change the password of %s",
+		         user, name));
 		fprintf(stderr,
 		        _("%s: %s is not authorized to change the password of %s\n"),
-		        Prog, user_context, name);
-		freecon(user_context);
-		exit(1);
+		        Prog, user, name);
+		if (NULL != user_context) {
+			freecon (user_context);
+		}
+		exit (E_NOPERM);
 	}
-#endif /* WITH_SELINUX */
+#endif				/* WITH_SELINUX */
 
 	/*
 	 * If the UID of the user does not match the current real UID,
@@ -1101,13 +1106,15 @@ int main (int argc, char **argv)
 	SYSLOG ((LOG_INFO, "password for '%s' changed by '%s'", name, myname));
 	closelog ();
 	if (!qflg) {
-		if (!eflg) {
-			puts (_("Password changed."));
+		if (!anyflag) {
+#ifndef USE_PAM
+			printf (_("%s: password changed."), Prog);
+#endif				/* USE_PAM */
 		} else {
-			puts (_("Password set to expire."));
+			printf (_("%s: password expiry information changed."), Prog);
 		}
 	}
-	exit (E_SUCCESS);
-	/* NOT REACHED */
+
+	return E_SUCCESS;
 }
 
