@@ -53,7 +53,7 @@
 
 #include <config.h>
 
-#ident "$Id: su.c 2666 2009-04-12 00:28:32Z nekral-guest $"
+#ident "$Id: su.c 2899 2009-05-10 18:20:41Z nekral-guest $"
 
 #include <getopt.h>
 #include <grp.h>
@@ -63,12 +63,14 @@
 #include <sys/types.h>
 #include "prototypes.h"
 #include "defines.h"
-#include "exitcodes.h"
 #include "pwauth.h"
 #include "getdef.h"
 #ifdef USE_PAM
 #include "pam_defs.h"
-#endif
+#endif				/* USE_PAM */
+/*@-exitarg@*/
+#include "exitcodes.h"
+
 /*
  * Assorted #defines to control su's behavior
  */
@@ -458,16 +460,16 @@ int main (int argc, char **argv)
 		}
 	}
 	if ('\0' == name[0]) {		/* use default user */
-		struct passwd *root_pw = getpwnam("root");
+		struct passwd *root_pw = getpwnam ("root");
 		if ((NULL != root_pw) && (0 == root_pw->pw_uid)) {
 			(void) strcpy (name, "root");
 		} else {
-			struct passwd *root_pw = getpwuid(0);
+			root_pw = getpwuid (0);
 			if (NULL == root_pw) {
-				SYSLOG((LOG_CRIT, "There is no UID 0 user."));
-				su_failure(tty);
+				SYSLOG ((LOG_CRIT, "There is no UID 0 user."));
+				su_failure (tty);
 			}
-			(void) strcpy(name, root_pw->pw_name);
+			(void) strcpy (name, root_pw->pw_name);
 		}
 	}
 
