@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: userdel.c 2691 2009-04-15 21:14:08Z nekral-guest $"
+#ident "$Id: userdel.c 2874 2009-05-09 13:15:44Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -56,16 +56,16 @@
 #include "pwauth.h"
 #include "pwio.h"
 #include "shadowio.h"
-#include "exitcodes.h"
 #ifdef	SHADOWGRP
 #include "sgroupio.h"
 #endif
+/*@-exitarg@*/
+#include "exitcodes.h"
+
 /*
  * exit status values
  */
-#define E_SUCCESS	0
 #define E_PW_UPDATE	1	/* can't update password file */
-#define E_USAGE		2	/* invalid command syntax */
 #define E_NOTFOUND	6	/* specified user doesn't exist */
 #define E_USER_BUSY	8	/* user currently logged in */
 #define E_GRP_UPDATE	10	/* can't update group file */
@@ -590,17 +590,17 @@ static void user_busy (const char *name, uid_t uid)
  * We see if the user is logged in by looking for the user name
  * in the utmp file.
  */
-#if HAVE_UTMPX_H
+#ifdef USE_UTMPX
 	struct utmpx *utent;
 
 	setutxent ();
 	while ((utent = getutxent ()) != NULL)
-#else
+#else				/* !USE_UTMPX */
 	struct utmp *utent;
 
 	setutent ();
 	while ((utent = getutent ()) != NULL)
-#endif
+#endif				/* !USE_UTMPX */
 	{
 		if (utent->ut_type != USER_PROCESS)
 			continue;
@@ -1021,7 +1021,6 @@ int main (int argc, char **argv)
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
-	exit ((0 != errors) ? E_HOMEDIR : E_SUCCESS);
-	/* NOT REACHED */
+	return ((0 != errors) ? E_HOMEDIR : E_SUCCESS);
 }
 
